﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Common;
using System.Drawing;
using HeuristicLab.Data;
using System.IO;
using HeuristicLab.Problems.Scheduling.Evaluators;
using HeuristicLab.Parameters;
using HeuristicLab.Problems.Scheduling.Analyzers;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Encodings.SchedulingEncoding;
using HeuristicLab.Encodings.SchedulingEncoding.Interfaces;
using HeuristicLab.Encodings.SchedulingEncoding.JobSequenceMatrix;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Encodings.SchedulingEncoding.PriorityRulesVector;
using HeuristicLab.Encodings.SchedulingEncoding.PermutationWithRepetition;
using HeuristicLab.Problems.Scheduling.Decoders;

namespace HeuristicLab.Problems.Scheduling {
  [Item("JobShop Scheduling Problem", "Represents a standard JobShop Scheduling Problem")]
  [Creatable("Problems")]
  [StorableClass]
  public sealed class JobShopSchedulingProblem : SchedulingProblem {
    [StorableConstructor]
    private JobShopSchedulingProblem(bool deserializing) : base(deserializing) { }
    private JobShopSchedulingProblem(JobShopSchedulingProblem original, Cloner cloner)
      : base(original, cloner) {
      this.encodingType = original.encodingType;
      this.Jobs = cloner.Clone(original.Jobs);
      this.EvaluatorType = cloner.Clone(original.EvaluatorType);
        AttachEventHandlers();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new JobShopSchedulingProblem(this, cloner);
    }


    #region Parameter Properties
    public ValueParameter<ItemList<Job>> JobDataParameter {
      get { return (ValueParameter<ItemList<Job>>)Parameters["JobData"]; }
    }
    public ValueParameter<SchedulingEvaluatorType> EvaluatorTypeParameter {
      get { return (ValueParameter<SchedulingEvaluatorType>)Parameters["EvaluatorType"]; }
    }
    public OptionalValueParameter<Schedule> BestKnownSolutionParameter {
      get { return (OptionalValueParameter<Schedule>)Parameters["BestKnownSolution"]; }
    }

    public ValueParameter<IntValue> JobsParameter {
      get { return (ValueParameter<IntValue>)Parameters["Jobs"]; }
    }
    public ValueParameter<IntValue> ResourcesParameter {
      get { return (ValueParameter<IntValue>)Parameters["Resources"]; }
    }

    #endregion

    #region Properties
    public ItemList<Job> JobData {
      get { return JobDataParameter.Value; }
      set { JobDataParameter.Value = value; }
    }
    public SchedulingEvaluatorType EvaluatorType {
      get { return EvaluatorTypeParameter.Value; }
      set { EvaluatorTypeParameter.Value = value; }
    }
    public Schedule BestKnownSolution {
      get { return BestKnownSolutionParameter.Value; }
      set { BestKnownSolutionParameter.Value = value; }
    }
    public IntValue Jobs {
      get { return JobsParameter.Value; }
      set { JobsParameter.Value = value; }
    }
    public IntValue Resources {
      get { return ResourcesParameter.Value; }
      set { ResourcesParameter.Value = value; }
    }
    #endregion

    [Storable]
    private SchedulingEncodingTypes encodingType;

    public JobShopSchedulingProblem()
      : base(new SchedulingEvaluationAlgorithm (), new JSMRandomCreator ()) {
      Parameters.Add(new ValueParameter<ItemList<Job>>("JobData", "Jobdata defining the precedence relationships and the duration of the tasks in this JSSP-Instance.", new ItemList<Job>()));
      Parameters.Add(new ValueParameter<SchedulingEvaluatorType>("EvaluatorType", "Type of the evaluator used to determine the quality of an individual.", new SchedulingEvaluatorType(SchedulingEvaluatorTypes.MakespanEvaluator)));
      Parameters.Add(new OptionalValueParameter<Schedule>("BestKnownSolution", "The best known solution of this JSSP instance."));
      
      Parameters.Add(new ValueParameter<IntValue>("Jobs", "The number of jobs used in this JSSP instance.", new IntValue ()));
      Parameters.Add(new ValueParameter<IntValue>("Resources", "The number of resources used this JSSP instance.", new IntValue()));
      encodingType = SchedulingEncodingTypes.JSMEncoding;
        InitializeOperators();
        AttachEventHandlers();
    }

    #region Events
    //TODO: Implement event handlers for problem specific events!
    protected override void OnSolutionCreatorChanged() {
      if (SolutionCreator.GetType().Equals(typeof(JSMRandomCreator)))
        encodingType = SchedulingEncodingTypes.JSMEncoding;
      else {
        if (SolutionCreator.GetType().Equals(typeof(PRVRandomCreator)))
          encodingType = SchedulingEncodingTypes.PRVEncoding;
        else
          encodingType = SchedulingEncodingTypes.PWREncoding;
      }
      InitializeOperators();
    }
    #endregion

    #region Helpers
    private void InitializeOperators() {
      Operators.Clear();
      ApplyEncoding();
      Operators.Add(new BestSchedulingSolutionAnalyzer());
      //ParameterizeAnalyzer();
      //ParameterizeOperators();
    }

    private void ApplyEncoding() {
      if (EvaluatorType.Value == SchedulingEvaluatorTypes.MakespanEvaluator) {
        if (encodingType == SchedulingEncodingTypes.JSMEncoding) {
          Operators.AddRange(ApplicationManager.Manager.GetInstances<IJSMOperator>());
          ((SchedulingEvaluationAlgorithm)this.EvaluatorParameter.ActualValue).InitializeOperatorGraph<JSMEncoding>(new JSMDecoder(), new MakespanEvaluator());
        } else {
          if (encodingType == SchedulingEncodingTypes.PRVEncoding) {
            Operators.AddRange(ApplicationManager.Manager.GetInstances<IPRVOperator>());
            ((SchedulingEvaluationAlgorithm)this.EvaluatorParameter.ActualValue).InitializeOperatorGraph<PRVEncoding>(new PRVDecoder(), new MakespanEvaluator());
          } else {
            if (encodingType == SchedulingEncodingTypes.PWREncoding) {
              Operators.AddRange(ApplicationManager.Manager.GetInstances<IPWROperator>());
              ((SchedulingEvaluationAlgorithm)this.EvaluatorParameter.ActualValue).InitializeOperatorGraph<PWREncoding>(new PWRDecoder(), new MakespanEvaluator());
            }
          }
        }
      } else {
        if (encodingType == SchedulingEncodingTypes.JSMEncoding) {
          Operators.AddRange(ApplicationManager.Manager.GetInstances<IJSMOperator>());
          ((SchedulingEvaluationAlgorithm)this.EvaluatorParameter.ActualValue).InitializeOperatorGraph<JSMEncoding>(new JSMDecoder(), new MeanTardinessEvaluator());
        } else {
          if (encodingType == SchedulingEncodingTypes.PRVEncoding) {
            Operators.AddRange(ApplicationManager.Manager.GetInstances<IPRVOperator>());
            ((SchedulingEvaluationAlgorithm)this.EvaluatorParameter.ActualValue).InitializeOperatorGraph<PRVEncoding>(new PRVDecoder(), new MeanTardinessEvaluator());
          } else {
            if (encodingType == SchedulingEncodingTypes.PWREncoding) {
              Operators.AddRange(ApplicationManager.Manager.GetInstances<IPWROperator>());
              ((SchedulingEvaluationAlgorithm)this.EvaluatorParameter.ActualValue).InitializeOperatorGraph<PWREncoding>(new PWRDecoder(), new MeanTardinessEvaluator());
            }
          }
        }
      }
    }

    private void AttachEventHandlers() {
      //TODO: Attach event handlers implemented aabove to appropriate events of the problem.
    }

    private void ParameterizeOperators() {
      throw new NotImplementedException();
    }

    private void ParameterizeAnalyzer() {
      throw new NotImplementedException();
    }
    #endregion

    #region Importmethods
    private List<string> SplitString(string line) {
      if (line == null)
        return null;
      List<string> data = new List<string>(line.Split(' '));
      List<string> result = new List<string>();

      foreach (string s in data) {
        if (!String.IsNullOrEmpty(s) && s != "" && s != " ")
          result.Add(s);
      }

      return result;
    }
    private int[] GetIntArray(List<string> data) {
      int[] arr = new int [data.Count];
      for (int i = 0; i < data.Count; i++ ) {
        arr[i] = Int32.Parse(data[i]);
      }
      return arr;
    }
    public void ImportFromORLibrary(string fileName) {
      if (!File.Exists(fileName))
        return;
      StreamReader problemFile = new StreamReader(fileName);
      //assures that the parser only reads the first problem instance given in the file
      bool problemFound = false;

      JobData = new ItemList<Job>();

      while (!problemFile.EndOfStream && !problemFound) {
        string line = problemFile.ReadLine();
        List<string> data = SplitString(line);
        if (data.Count > 0 && ((int)data[0][0] >= 48 && (int)data[0][0] <= 57)) {
          int jobCount = 0;
          Jobs = new IntValue (Int32.Parse(data[0]));
          Resources = new IntValue (Int32.Parse(data[1]));
          //data[2] = bestKnownQuality (double)
          //data[3] = dueDates (0|1)
          bool withDueDates = false;
          if (data.Count > 2)
            BestKnownQualityParameter.ActualValue = new DoubleValue(Double.Parse (data[2]));
          if (data.Count > 3 && data[3] == "1")
            withDueDates = true;
          line = problemFile.ReadLine();
          data = SplitString(line);
          while (!problemFile.EndOfStream && data.Count > 0 && ((int)data[0][0] >= 48 && (int)data[0][0] <= 57)) {
            DoubleValue dueDate = null;
            int dataCount = data.Count;
            if (withDueDates) {
              dueDate = new DoubleValue(Double.Parse(data[data.Count - 1]));
              dataCount--;
            }
            Job j = new Job(new IntValue(jobCount), dueDate);
            for (int i = 0; i < dataCount; i++) {
              Task t = new Task(i / 2, Int32.Parse(data[i]), jobCount, Double.Parse(data[i + 1]));
              j.Tasks.Add(t);
              i++;
            }//for
            this.JobData.Add(j);
            jobCount++;
            line = problemFile.ReadLine();
            data = SplitString(line);
          }//while
          problemFound = true;
        }//if
      }//while
      problemFile.Close();
    }

    public void ImportJSMSolution(string fileName) {
      if (!File.Exists(fileName))
        return;
      StreamReader solutionFile = new StreamReader(fileName);
      //assures that the parser only reads the first solution instance given in the file
      bool solutionFound = false;
      JSMEncoding solution = new JSMEncoding();
      while (!solutionFile.EndOfStream && !solutionFound) {
        
        string line = solutionFile.ReadLine();
        List<string> data = SplitString(line);
        if (data.Count > 0 && ((int)data[0][0] >= 48 && (int)data[0][0] <= 57)) {
          if (data.Count > 2)
            BestKnownQualityParameter.ActualValue = new DoubleValue(Double.Parse(data[2]));
          line = solutionFile.ReadLine();
          data = SplitString(line);
          while (data != null && data.Count > 0 && ((int)data[0][0] >= 48 && (int)data[0][0] <= 57)) {
            Permutation p = new Permutation(PermutationTypes.Absolute, GetIntArray (data));
            solution.JobSequenceMatrix.Add(p);

            line = solutionFile.ReadLine();
            data = SplitString(line);
          }//while
          solutionFound = true;
        }//if
      }//while
      solutionFile.Close();

      JSMDecoder decoder = new JSMDecoder();
      Schedule result = decoder.CreateScheduleFromEncoding(solution, JobData);
      BestKnownSolution = result;
    }
    #endregion

  }
}
